#!/usr/bin/env python

import argparse
import datetime
import logging
import numpy as np
import os
from PIL import Image
import subprocess


__author__ = 'Steve Marple'
__version__ = '0.0.11'
__license__ = 'MIT'


def find_start_end_frames(st, et, step, filename_fstr):
    r_st = r_et = None
    t = st
    while t < et:
        dt = datetime.datetime.fromtimestamp(t)
        filename = filename_fstr.format(DateTime=dt)
        if os.path.exists(filename):
            r_st = t
            break
        t += step
    if r_st is None:
        return None, None

    t = et
    while t >= r_st:
        dt = datetime.datetime.fromtimestamp(t)
        filename = filename_fstr.format(DateTime=dt)
        if os.path.exists(filename):
            r_et = t
            break
        t -= step

    return r_st, r_et


def ffmpeg(start_time, end_time, step, filename_fstr, output_filename,
           resolution=None, duration=None, speed_up=None, ifr=None, ofr=None):
    img_mode = 'RGB'  # Mode for all images
    # Remember et is inclusive
    st, et = find_start_end_frames(start_time.astype('datetime64[s]').astype(int),
                                   end_time.astype('datetime64[s]').astype(int),
                                   step,
                                   filename_fstr)
    if st is None:
        raise ValueError('could not find any images in time range')

    input_duration = et - st
    frames = input_duration / step  # Includes repeated frames
    if ifr is not None:
        pass
    elif duration is not None:
        ifr = round(float(frames) / duration)
    elif speed_up is not None:
        ifr = speed_up * float(frames) / input_duration
    elif ifr is None:
        ifr = float(frames) / input_duration

    if ofr is None:
        ofr = 60
        # Can a lower sensible rate be used, but one which is still larger than ifr?
        for fr in (5, 10, 15, 20, 30, 60):
            if fr >= ifr:
                ofr = fr
                break

    first_frame = Image.open(filename_fstr.format(DateTime=datetime.datetime.fromtimestamp(st))).convert('RGB')

    # TODO: Add code to run tasks at start/end of time lapse, eg add text, dissolving to/from first/last frames
    if resolution is None:
        resolution = tuple(first_frame.size)
    else:
        resolution = tuple(resolution)  # Force to be tuple so that img.size comparison works later

    # Set up a subprocess to run ffmpeg
    cmd = ('ffmpeg',
           '-loglevel', 'error',
           '-y',  # overwrite
           '-framerate', str(ifr),  # input frame rate
           '-s', '%dx%d' % (resolution[0], resolution[1]),  # size of image
           '-pix_fmt', 'rgb24',  # format
           '-f', 'rawvideo',
           '-i', '-',  # read from stdin
           '-vcodec', 'libx264',  # set output encoding
           '-r', str(ofr),  # output frame rate
           output_filename)
    logger.info('Running command ' + ' '.join(cmd))
    proc = subprocess.Popen(cmd, stdin=subprocess.PIPE)

    img = None
    t = st
    while t <= et:
        dt = datetime.datetime.fromtimestamp(t)
        filename = filename_fstr.format(DateTime=dt)
        if os.path.exists(filename):
            logger.debug('reading %s', filename)
            img = Image.open(filename)
        elif img is None:
            logger.debug('missing %s', filename)
            img = Image.new('RGB', resolution)
        else:
            logger.debug('missing intermediate image %s', filename)

        if img.mode != img_mode:
            logger.debug('converting image to mode %s', img_mode)
            img = img.convert(img_mode)
        if tuple(img.size) != resolution:
            logger.debug('resizing image')
            img = img.resize(resolution, Image.BILINEAR)

        proc.stdin.write(img.tobytes())
        t += step
    proc.communicate()
    logger.info('saved to %s', output_filename)


logger = logging.getLogger(__name__)
if __name__ == '__main__':
    default_config_file = os.path.join(os.path.sep, 'etc', 'camera.ini')
    parser = \
        argparse.ArgumentParser(description='')

    parser.add_argument('-c', '--config-file',
                        default=default_config_file,
                        help='Configuration file')

    parser.add_argument('-e', '--end-time',
                        required=True,
                        help='End time for time lapse')
    parser.add_argument('-f', '--fstr',
                        required=True,
                        help='Filenames format string')
    parser.add_argument('--log-level',
                        choices=['debug', 'info', 'warning',
                                 'error', 'critical'],
                        default='info',
                        help='Control how much detail is printed',
                        metavar='LEVEL')
    parser.add_argument('-o', '--output',
                        required=True,
                        help='Output filename')
    parser.add_argument('-r', '--output-frame-rate',
                        type=float,
                        help='Input frame rate',
                        metavar='FPS')
    parser.add_argument('--resolution',
                        nargs=2,
                        type=int,
                        help='Video resolution',
                        metavar=('WIDTH', 'HEIGHT'))
    parser.add_argument('-s', '--start-time',
                        required=True,
                        help='Start time for time lapse')
    parser.add_argument('--step',
                        default=1,
                        type=float,
                        help='Time step between images',
                        metavar='SECONDS')

    # Offer multiple methods to set the input frame rate
    ifr_group = parser.add_mutually_exclusive_group()
    ifr_group.add_argument('-d', '--duration',
                           type=float,
                           help='Set output duration for timelapse',
                           metavar='SECONDS')
    ifr_group.add_argument('-i', '--input-frame-rate',
                           type=float,
                           help='Input frame rate',
                           metavar='FPS')
    ifr_group.add_argument('--speed-up',
                           type=float,
                           help='Playback speed as ratio to real time',
                           metavar='FPS')

    args = parser.parse_args()

    logging.basicConfig(level=getattr(logging, args.log_level.upper()))

    start_time = np.datetime64(args.start_time, 's')
    end_time = np.datetime64(args.end_time, 's')

    kwargs = {}
    if args.input_frame_rate is not None:
        kwargs['ifr'] = args.input_frame_rate
    elif args.speed_up is not None:
        kwargs['speed_up'] = args.speed_up
    else:
        kwargs['duration'] = args.duration

    ffmpeg(start_time, end_time, args.step, args.fstr, args.output,
           ofr=args.output_frame_rate,
           resolution=args.resolution,
           **kwargs)



from fractions import Fraction
import logging
from picamera import PiCamera, PiCameraNotRecording
from picamera.array import PiRGBArray
from PIL import Image
import threading
import time
import traceback

from cameralogger import get_config_option


__author__ = 'Steve Marple'
__version__ = '0.0.11'
__license__ = 'MIT'


class Camera(object):
    def __init__(self, config):
        self.config = config
        self.capture_image_lock = threading.Lock()
        self.camera = PiCamera()
        self.use_video_port = False
        self.splitter_port = None

    def __del__(self):
        self.stop_recording()
        self.camera.close()

    def apply_settings(self, section):
        # Stop any previous recordings
        self.stop_recording()

        self.use_video_port = get_config_option(self.config, section, 'use_video_port', get='getboolean')
        if self.use_video_port:
            self.splitter_port = get_config_option(self.config, section, 'splitter_port', 1, get='getint')

        # Fractions
        framerate = get_config_option(self.config, section, 'framerate')
        if framerate is not None:
            logger.debug('setting framerate=%s', str(framerate))
            self.camera.framerate = fraction_or_float(framerate)

        # Booleans
        for k in ('hflip', 'image_denoise', 'vflip', 'video_denoise'):
            val = get_config_option(self.config, section, k, get='getboolean')
            if val is not None:
                logger.debug('setting %s=%s', k, 'true' if val else 'false')
                setattr(self.camera, k, val)

        # Ints
        for k in ('brightness', 'contrast', 'exposure_compensation', 'iso', 'rotation', 'saturation', 'sensor_mode',
                  'sharpness', 'shutter_speed'):
            val = get_config_option(self.config, section, k, get='getint')
            if val is not None:
                logger.debug('setting %s=%d', k, val)
                setattr(self.camera, k, val)

        # Strings
        for k in ('awb_mode', 'drc_strength', 'exposure_mode', 'meter_mode', 'resolution', 'still_stats'):
            val = get_config_option(self.config, section, k)
            if val is not None:
                logger.debug('setting %s=%s', k, val)
                setattr(self.camera, k, val)

        awb_gains = get_config_option(self.config, section, 'awb_gains')
        if awb_gains:
            awb_gains = awb_gains.split()
            if len(awb_gains) == 1:
                self.camera.awb_gains = fraction_or_float(awb_gains[0])
            else:
                # awb_gains = (Fraction(*map(int, awb_gains[0])), Fraction(*map(int, awb_gains[0])))
                self.camera.awb_gains = (fraction_or_float(awb_gains[0]), fraction_or_float(awb_gains[1]))

        if self.use_video_port:
            # Enable recording
            try:
                self.camera.start_recording(NullStream(), format='rgb', splitter_port=self.splitter_port)
            finally:
                self.stop_recording()

    def capture_image(self, section):
        logger.debug('capture_image: acquiring lock')
        if self.capture_image_lock.acquire(False):
            try:
                logger.debug('capture_image: acquired lock')
                t = time.time()
                if self.use_video_port and not self.camera.frame.complete:
                    # With long exposures the camera has not started producing valid frames. Attempting to
                    # record now mean the thread hangs and prevents other captures later.
                    raise Exception('no video frame to capture')

                shape = list(self.camera.resolution)
                shape.append(3)
                data = PiRGBArray(self.camera)
                self.camera.capture(data,
                                    format='rgb',
                                    use_video_port=self.use_video_port,
                                    splitter_port=self.splitter_port)
                img = Image.fromarray(data.array)
                img_info = self.get_image_settings(t)

                return img, img_info, t

            finally:
                logging.debug('capture_image: released lock')
                self.capture_image_lock.release()
        else:
            logger.warning('capture_image: could not acquire lock')
            raise Exception('could not acquire lock')

    def get_image_settings(self, t):
        r = {'Exposure_s': self.camera.exposure_speed / 1000000.0,
             'ExposureMode': self.camera.exposure_mode,
             'SystemTemperature': float('NaN'),
             'ISO': int(self.camera.iso),
             'DigitalGain': float(self.camera.digital_gain),
             'AnalogGain': float(self.camera.analog_gain),
             'SensorMode': int(self.camera.sensor_mode),
             }
        # Take CPU temperature as system temperature
        try:
            with open('/sys/class/thermal/thermal_zone0/temp') as f:
                r['SystemTemperature'] = float(f.read().strip()) / 1000
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            logger.warning('could not read system temperature')
            logger.debug(traceback.format_exc())
        return r

    def stop_recording(self):
        # Stop any previous video recording. How to tell if this has to be done? It is not clear to
        # which splitter port camera.recording refers.
        if self.use_video_port:
            try:
                self.camera.stop_recording(self.splitter_port)
            except PiCameraNotRecording:
                pass
            finally:
                self.splitter_port = None
                self.use_video_port = False



class NullStream(object):
    def __init__(self):
        pass

    def write(self, _):
        pass


def fraction_or_float(s):
    if '/' in s:
        return Fraction(*map(int, s.split('/')))
    else:
        return float(s)


logger = logging.getLogger(__name__)



#!/usr/bin/env python

import argparse
import importlib
import logging
import os
import signal
import sys
import threading
import time
import traceback

import cameralogger


__author__ = 'Steve Marple'
__version__ = '0.0.11'
__license__ = 'MIT'


def run_camera(forced_schedule):
    global config
    global camera
    global sampling_interval

    # This should be called after dropping root privileges because
    # it uses safe_eval to convert strings to numbers or
    # lists (not guaranteed safe!)
    #camera = cameralogger.get_camera(config)
    #cameralogger.init_camera(camera, config)

    camera_type = config.get('camera', 'type')
    logger.debug('camera type: %s', camera_type)

    Camera = getattr(importlib.import_module('cameralogger.' + camera_type), 'Camera')
    camera = Camera(config)
    camera.apply_settings('camera')

    signal.signal(signal.SIGTERM, stop_handler)
    signal.signal(signal.SIGINT, stop_handler)

    try:
        get_log_file_for_time(time.time(), log_filename)
        logger.info('Starting sampling thread')

        do_every(camera, config, forced_schedule, cameralogger.process_tasks)
        while take_images:
            time.sleep(2)

        # Wait until all other threads have (or should have)
        # completed
        try:
            sampling_interval_lock.acquire(True)
            si = sampling_interval
        finally:
            sampling_interval_lock.release()

        for n in range(int(round(si)) + 1):
            if threading.activeCount() == 1:
                break
            time.sleep(1)
    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception:
        get_log_file_for_time(time.time(), log_filename)
        logger.error(traceback.format_exc())
        time.sleep(5)


def timeout(func, args=(), kwargs={}, timeout_duration=1, default=None):
    class TimeoutError(Exception):
        pass

    def handler(signum, frame):
        raise TimeoutError()

    # set the timeout handler
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(timeout_duration)
    try:
        result = func(*args, **kwargs)
    except TimeoutError as exc:
        result = default
    finally:
        signal.alarm(0)

    return result


def cancel_sampling_threads():
    threads = threading.enumerate()
    for t in threads[1:]:
        t.cancel()

        # If all the other threads have completed then exit; exit anyway
    # after a short time has passed
    t = time.time()
    while time.time() < t + 1 and len(threading.enumerate()) > 2:
        time.sleep(0.1)
        # sys.exit()


take_images = True


def stop_handler(signal, frame):
    global take_images
    global camera
    get_log_file_for_time(time.time(), log_filename)
    logger.info('Stopping sampling threads')
    take_images = False
    cancel_sampling_threads()
    if camera:
        camera = None


def do_every(camera, config, forced_schedule, worker_func, iterations=0):
    global sampling_interval
    if iterations != 1:
        # Identify current operating condition to find the actions to do and sampling interval
        schedule, schedule_info = cameralogger.get_schedule(config, forced_schedule)
        logger.info('using schedule %s', schedule)
        logger.info('schedule info: %s', repr(schedule_info))
        # Schedule the next worker thread. Aim to start at the next
        # multiple of sampling interval. Take current time, add 1.25
        # of the interval and then find the nearest
        # multiple. Calculate delay required.
        now = time.time()
        interval = cameralogger.get_config_option(config, schedule, 'sampling_interval',
                                                  fallback_section='common',
                                                  default=default_sampling_interval,
                                                  get='getfloat')
        delay = round_to(now + (1.25 * interval), interval) - now
        # Avoid lockups by many threads piling up. Impose a minimum delay
        if delay < 0.1:
            delay = 0.1
        t = threading.Timer(delay,
                            do_every,
                            [camera, config, forced_schedule, worker_func,
                             0 if iterations == 0 else iterations - 1])
        t.daemon = True
        t.start()

        # Check if the schedule has changed, if so apply any new settings. To access last_schedule and camera
        # the camera_lock must be acquired
        if config.has_option(schedule, 'camera_settings'):
            if camera_lock.acquire(False):
                try:
                    global last_schedule
                    logger.debug('camera_lock: acquired lock')
                    if schedule != last_schedule:
                        settings_section = config.get(schedule, 'camera_settings')
                        if not config.has_section(settings_section):
                            logger.error('cannot apply camera settings for schedule %s: section %s does not exist',
                                         schedule, settings_section)
                            raise Exception('missing section %s' % settings_section)

                        if last_schedule is not None and settings_section == config.get(last_schedule,
                                                                                        'camera_settings'):
                            # No action needed
                            logger.debug('schedule changed but camera settings unchanged')
                        else:
                            logger.debug('schedule changed, applying new settings for %s from section %s',                                 schedule, settings_section)
                            camera.apply_settings(settings_section)
                        last_schedule = schedule  # Update last, after settings have been applied
                finally:
                    camera_lock.release()
                    logging.debug('camera_lock: released lock')
            else:
                logger.error('camera_lock: could not acquire lock')

        # Update the global sampling_interval so that cancel_sampling_threads knows how long to wait.
        # This could be attempted by multiple capture threads so must use a lock. Don't wait for the lock to be
        # available.
        if sampling_interval_lock.acquire(False):
            try:
                logger.debug('sampling_interval_lock: acquired lock')
                sampling_interval = interval
            finally:
                logging.debug('sampling_interval_lock: released lock')
                sampling_interval_lock.release()
        else:
            logger.error('sampling_interval_lock: could not acquire lock')

    try:
        worker_func(camera, config, schedule, schedule_info)
    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception:
        get_log_file_for_time(time.time(), log_filename)
        logger.error(traceback.format_exc())


def round_to(n, nearest):
    return round(n / float(nearest)) * nearest


def get_log_file_for_time(t, fstr,
                          mode='a',
                          delay=True,
                          name=__name__):
    if fstr is None:
        return
    fh = get_log_file_for_time.fh
    tmp_name = time.strftime(fstr, time.gmtime(t))

    if fh is not None:
        if fh.stream and fh.stream.closed:
            fh = None
        elif fh.stream.name != tmp_name:
            # Filename has changed
            fh.close()
            fh = None

    if fh is None:
        # File wasn't open or filename changed
        p = os.path.dirname(tmp_name)
        if not os.path.isdir(p):
            os.makedirs(p)

        fh = logging.FileHandler(tmp_name, mode=mode, delay=delay)
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s',
                                      datefmt='%Y-%m-%dT%H:%M:%SZ')
        fh.setFormatter(formatter)
        log = logging.getLogger(name)
        for h in log.handlers:
            # Only remove file handlers 
            if isinstance(h, logging.FileHandler):
                log.removeHandler(h)
        log.addHandler(fh)


get_log_file_for_time.fh = None

camera_lock = threading.Lock()
# Access to camera and last_schedule controlled by camera_lock
camera = None
last_schedule = None

default_sampling_interval = 120

sampling_interval_lock = threading.Lock()
# Access to sampling_interval controlled by sampling_interval_lock
sampling_interval = default_sampling_interval

if __name__ == '__main__':

    logger = logging.getLogger(__name__)
    # Parse command line arguments
    progname = os.path.basename(sys.argv[0]).partition('.')[0]
    default_config_file = \
        os.path.join(os.path.sep, 'etc', 'camera.ini')

    parser = \
        argparse.ArgumentParser(description='Process and save images from a camera')

    parser.add_argument('-c', '--config-file',
                        default=default_config_file,
                        help='Configuration file')
    parser.add_argument('--log-level',
                        choices=['debug', 'info', 'warning',
                                 'error', 'critical'],
                        default='info',
                        help='Control how much detail is printed',
                        metavar='LEVEL')
    parser.add_argument('--log-format',
                        default='%(levelname)s:%(message)s',
                        help='Set format of log messages',
                        metavar='FORMAT')
    parser.add_argument('--schedule',
                        help='Override automatic scheduling')

    args = parser.parse_args()

    config = cameralogger.read_config_file(args.config_file)
    logging.basicConfig(level=getattr(logging, args.log_level.upper()),
                        format=args.log_format, datefmt='%Y-%m-%dT%H:%M:%SZ')

    if (args.schedule is not None and
            (not config.has_section(args.schedule) or not config.has_option(args.schedule, 'sampling_interval'))):
        raise Exception('%s is not a valid schedule' % args.schedule)

    log_filename = None
    if config.has_option('logfile', 'filename'):
        log_filename = config.get('logfile', 'filename')

    get_log_file_for_time(time.time(), log_filename)
    logger.info(progname + ' started')

    run_camera(args.schedule)


logger = logging.getLogger(__name__)



import logging
from PIL import Image
import threading
import time
import zwoasi
from cameralogger import get_config_option


__author__ = 'Steve Marple'
__version__ = '0.0.11'
__license__ = 'MIT'


class Camera(object):
    def __init__(self, config):
        num_cameras = zwoasi.get_num_cameras()
        if num_cameras == 0:
            raise Exception('no camera present')
        self.camera = zwoasi.Camera(0)
        self.config = config
        self.capture_image_lock = threading.Lock()

    def __del__(self):
        self.camera.stop_video_capture()
        self.camera.close()

    def apply_settings(self, section):
        # Initialise
        controls = self.camera.get_controls()
        self.camera.start_video_capture()

        # Read all camera controls defined in the config file
        for c in controls:
            cl = c.lower()
            value = get_config_option(self.config, section, cl)
            if value is not None:
                default_value = controls[c]['DefaultValue']
                control_type = getattr(zwoasi, 'ASI_' + c.upper())
                logger.debug('set control value %s to %s', cl, value)
                if value == 'auto':
                    self.camera.set_control_value(control_type, default_value, auto=True)
                else:
                    # Cast value to same type as default_value
                    self.camera.set_control_value(control_type, type(default_value)(value), auto=False)

        image_type = get_config_option(self.config, section, 'image_type')
        if image_type is not None:
            logger.debug('set image type to %s', image_type)
            self.camera.set_image_type(getattr(zwoasi, 'ASI_IMG_' + image_type.upper()))

    def capture_image(self, _):
        logger.debug('capture_image: acquiring lock')
        if self.capture_image_lock.acquire(False):
            try:
                logger.debug('capture_image: acquired lock')
                t = time.time()
                img_info = self.get_control_values()
                img = Image.fromarray(self.camera.capture_video_frame()[:, :, ::-1])  # Swap from BGR order
                img_info['DateTime'] = time.strftime('%Y-%m-%d %H:%M:%S+00:00', time.gmtime(t))

                # Take CPU temperature as system temperature
                img_info['SystemTemperature'] = float('NaN')
                with open('/sys/class/thermal/thermal_zone0/temp') as f:
                    img_info['SystemTemperature'] = float(f.read().strip()) / 1000
                return img, img_info, t

            finally:
                logging.debug('capture_image: released lock')
                self.capture_image_lock.release()
        else:
            logger.warning('capture_image: could not acquire lock')
            raise Exception('could not acquire lock')

    def get_control_values(self):
        controls = self.camera.get_controls()
        r = {}
        for k in controls:
            r[k] = self.camera.get_control_value(controls[k]['ControlType'])[0]

        # Fix up certain keys
        r['Exposure_s'] = r['Exposure'] / 1000000.0
        if 'Temperature' in r:
            r['SensorTemperature'] = r['Temperature'] / 10.0
        if 'Flip' in r:
            r['Flip'] = {0: 'None', 1: 'Horizontal', 2: 'Vertical', 3: 'Both'}[r['Flip']]

        return r


logger = logging.getLogger(__name__)



import logging
import glob
from PIL import Image
import threading
import time
from cameralogger import get_config_option


__author__ = 'Steve Marple'
__version__ = '0.0.11'
__license__ = 'MIT'


class Camera(object):
    def __init__(self, config):
        self.config = config
        self.file_gen = None
        self.capture_image_lock = threading.Lock()

    def apply_settings(self, section):
        self.file_gen = self.get_next_file(section)

    def capture_image(self, _):
        logger.debug('capture_image: acquiring lock')
        if self.capture_image_lock.acquire(False):
            try:
                logger.debug('capture_image: acquired lock')
                t = time.time()
                filename = next(self.file_gen)
                if filename is None:
                    raise Exception('no file to read')
                logger.debug('loading image from %s', filename)
                img = Image.open(filename)
                # Fake some image details
                img_info = dict(Exposure_s=1.0,
                                Gain=1,
                                Aperture=2.0,
                                DateTime=time.strftime('%Y-%m-%d %H:%M:%S+00:00', time.gmtime(t)),
                                SensorTemperature=25.0,
                                SystemTemperature=float('NaN'))

                # Take CPU temperature as system temperature
                img_info['SystemTemperature'] = float('NaN')
                with open('/sys/class/thermal/thermal_zone0/temp') as f:
                    img_info['SystemTemperature'] = float(f.read().strip()) / 1000
                return img, img_info, t

            finally:
                self.capture_image_lock.release()
                logging.debug('capture_image: released lock')
        else:
            logger.warning('capture_image: could not acquire lock')
            raise Exception('could not acquire lock')

    def get_next_file(self, section):
        while True:
            pattern = get_config_option(self.config, section, 'images')
            files = glob.glob(pattern)
            if not files:
                # raise Exception('no files matched %s', pattern)
                yield None
            for f in sorted(files):
                yield f


logger = logging.getLogger(__name__)



#!/usr/bin/env python

import codecs
import glob
import os
import re

from setuptools import setup, find_packages


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, *parts), 'rb', 'utf-8') as f:
        return f.read()


def find_version(*file_paths):
    """
    Build a path from *file_paths* and search for a ``__version__``
    string inside.
    """
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

print('==================================')
print(find_packages(exclude=['tests*']))
print('==================================')

setup(
    name='cameralogger',
    version=find_version('cameralogger/__init__.py'),
    description='Tools to record and decorate camera images for timelapses etc.',
    long_description=(read('README.rst') + '\n\n' +
                      read('AUTHORS.rst')),
    url='https://github.com/stevemarple/python-cameralogger',
    license='PSF',
    author='Steve Marple',
    author_email='s.marple@lancaster.ac.uk',
    packages=find_packages(exclude=['tests*']),
    install_requires=['atomiccreate', 'matplotlib', 'numpy>=1.7', 'PIL', 'scipy'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    scripts=glob.glob('cameralogger/bin/*.py'),
)




import astral
from atomiccreate import atomic_symlink
from atomiccreate import smart_open
import datetime
from fractions import Fraction
import logging
import numpy as np
import operator
import os
from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
from PIL import ImageOps
import re
import six
import subprocess
import sys
import time
import traceback


from cameralogger.formatters import MyFormatter
from cameralogger.formatters import LatLon
from aurorawatchuk.snapshot import AuroraWatchUK as  AuroraWatchUK_SS

if sys.version_info[0] >= 3:
    # noinspection PyCompatibility
    from configparser import RawConfigParser
else:
    # noinspection PyCompatibility
    from ConfigParser import RawConfigParser

__author__ = 'Steve Marple'
__version__ = '0.0.11'
__license__ = 'MIT'


class Tasks(object):
    """A collection of tasks than can be performed to create, act upon and save image buffers."""

    def __init__(self, camera=None, config=None, schedule=None, schedule_info={}):
        self.camera = camera
        self.config = config
        self.schedule = schedule
        self.schedule_info = schedule_info
        self.buffers = {}
        self.time = None
        self.capture_info = None
        self.format_dict = None

    def _get_color(self, section, default=None, fallback_section=None, get=None, raise_=True, option='color'):
        color = self._get_option(section, option, default=default, fallback_section=fallback_section, get=get,
                                 raise_=raise_)
        if re.match('^[0-9a-f]{3,6}$', color, re.IGNORECASE):
            return '#' + color
        else:
            return color

    def _get_mode(self, section, default=None, fallback_section=None, get=None, raise_=True, option='mode'):
        mode = self._get_option(section, option, default=default, fallback_section=fallback_section, get=get,
                                raise_=raise_)
        if len(mode) and mode[0] == '@':
            mode = mode.lstrip('@')
            if mode not in self.buffers:
                raise ValueError('no buffer named "%s"' % mode)
            return self.buffers[mode].mode
        else:
            return mode

    def _get_option(self, section, option, default=None, fallback_section=None, get=None, raise_=True):
        r = get_config_option(self.config, section, option, default=default, fallback_section=fallback_section, get=get)
        if raise_ and r is None:
            if fallback_section is None:
                raise Exception('could not find value for "%s" in section "%s"' % (option, section))
            else:
                raise Exception(
                    'could not find value for "%s" in sections "%s" or "%s"' % (option, section, fallback_section))
        return r

    def _get_size(self, section, default=None, fallback_section=None, get=None, raise_=True, option='size'):
        size = self._get_option(section, option, default=default, fallback_section=fallback_section, get=get,
                                raise_=raise_)
        if len(size) and size[0] == '@':
            size = size.lstrip('@')
            if size not in self.buffers:
                raise ValueError('no buffer named "%s"' % size)
            return self.buffers[size].size
        else:
            return map(int, size.split())

    def _make_dict(self, section):
        if self.format_dict is not None:
            return self.format_dict

        d = {}
        if self.capture_info is not None:
            # Copy all exposure settings
            d = self.capture_info.copy()
            # Remove subsecond part from time (datetime %S does not work as expected)
            d['DateTime'] = datetime.datetime.utcfromtimestamp(int(self.time))
        d['Schedule'] = self.schedule
        d['Section'] = section
        lat = self._get_option(section, 'latitude', fallback_section='common', get='getfloat')
        lon = self._get_option(section, 'longitude', fallback_section='common', get='getfloat')
        lat_lon = LatLon(lat, lon)
        d['LatLon'] = lat_lon

        for k, v in six.iteritems(self.schedule_info):
            d[k] = v
        self.format_dict = d
        return self.format_dict

    def format_str(self, section, s):
        return MyFormatter().format(s, **self._make_dict(section))

    def run_tasks(self, tasks):
        for section in tasks:
            if not self.config.has_section(section):
                raise ValueError('no section called %s' % section)
            act = self.config.get(section, 'action')
            if not hasattr(self, act):
                raise ValueError('unknown action (%s)' % act)
            logger.debug('calling action %s(%s)' % (act, repr(section)))
            getattr(self, act)(section)

    # Tasks are below
    def add_text(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        text = self._get_option(section, 'text')
        font_name = self._get_option(section, 'font', fallback_section='common')
        font_size = self._get_option(section, 'fontsize', fallback_section='common', get='getint')
        color = self._get_color(section, fallback_section='common')
        unicode = self._get_option(section, 'unicode', False, fallback_section='common', get='getboolean')
        position = map(int, self._get_option(section, 'position').split())
        font = ImageFont.truetype(font_name, font_size)
        if unicode:
            text = unescape_unicode(text)
        text = self.format_str(section, text)
        color = self.format_str(section, color)
        if dst != src:
            self.buffers[dst] = self.buffers[src].copy()
        draw = ImageDraw.Draw(self.buffers[dst])
        draw.text(position, text, color, font=font)

    def add_multiline_text(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        text = self._get_option(section, 'text')
        font_name = self._get_option(section, 'font', fallback_section='common')
        font_size = self._get_option(section, 'fontsize', fallback_section='common', get='getint')
        color = self._get_color(section, fallback_section='common')
        unicode = self._get_option(section, 'unicode', False, fallback_section='common', get='getboolean')
        spacing = self._get_option(section, 'spacing', fallback_section='common', get='getint')
        align = self._get_option(section, 'spacing', 'left', fallback_section='common')
        position = list(map(int, self._get_option(section, 'position').split()))
        font = ImageFont.truetype(font_name, font_size)
        if unicode:
            text = unescape_unicode(text)
        text = self.format_str(section, text)
        color = self.format_str(section, color)
        if dst != src:
            self.buffers[dst] = self.buffers[src].copy()
        draw = ImageDraw.Draw(self.buffers[dst])
        if hasattr(draw, 'multiline_text'):
            draw.multiline_text(position, text, color, font=font, spacing=spacing, align=align)
        else:
            # Compatibility, but without align
            for s in text.splitlines():
                draw.text(position, s, color, font=font)
                position[1] += spacing

    def alpha_composite(self, section):
        src1 = self._get_option(section, 'src1')
        src2 = self._get_option(section, 'src2')
        dst = self._get_option(section, 'dst', src1)
        self.buffers[dst] = Image.alpha_composite(self.buffers[src1], self.buffers[src2])

    def autocontrast(self, section):
        src = self._get_option(section, 'src')
        cutoff = self._get_option(section, 'cutoff', 0, get='getfloat')
        ignore = self._get_option(section, 'ignore', raise_=False)
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = ImageOps.autocontrast(self.buffers[src], cutoff, ignore)

    def blend(self, section):
        src1 = self._get_option(section, 'src1')
        src2 = self._get_option(section, 'src2')
        dst = self._get_option(section, 'dst', src1)
        alpha = self._get_option(section, 'alpha', get='getfloat')
        self.buffers[dst] = Image.blend(self.buffers[src1], self.buffers[src2], alpha)

    def capture(self, section):
        dst = self._get_option(section, 'dst')
        img, info, t = self.camera.capture_image(section)
        if self.time is None:
            self.time = t  # Record time of first capture
            self.capture_info = info
        self.buffers[dst] = img

    def colorize(self, section):
        src = self._get_option(section, 'src')
        black = self._get_option(section, 'black')
        white = self._get_option(section, 'white')
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = ImageOps.colorize(self.buffers[src], black, white)

    def command(self, section):
        """Run shell command."""
        cmd = self._get_option(section, 'cmd')
        check_call = self._get_option(section, 'check_call', True, get='getboolean')
        background = self._get_option(section, 'background', False, get='getboolean')
        cmd = self.format_str(section, cmd)
        if check_call:
            logger.debug('running command "%s"', cmd)
            subprocess.check_call(cmd, shell=True)
        elif background:
            logger.debug('running command "%s" in background', cmd)
            raise Exception('not implemented')
        else:
            logger.debug('running command "%s" (no checks)', cmd)
            subprocess.call(cmd, shell=True)

    def composite(self, section):
        src1 = self._get_option(section, 'src1')
        src2 = self._get_option(section, 'src2')
        mask = self._get_option(section, 'mask')
        dst = self._get_option(section, 'dst', src1)
        self.buffers[dst] = Image.composite(self.buffers[src1], self.buffers[src2], self.buffers[mask])

    def convert(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        mode = self._get_mode(section)
        self.buffers[dst] = self.buffers[src].convert(mode)

    def copy(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst')
        self.buffers[dst] = self.buffers[src].copy()

    def crop(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        position = map(int, self._get_option(section, 'position').split())
        self.buffers[dst] = self.buffers[src].crop(position)
        self.buffers[dst].load()  # crop() is lazy operation; break connection

    def delete(self, section):
        src = self._get_option(section, 'src')
        del self.buffers[src]

    def equalize(self, section):
        src = self._get_option(section, 'src')
        mask = self._get_option(section, 'mask', raise_=False)
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = ImageOps.equalize(self.buffers[src], mask)

    def expand(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        # border is left, top, right, bottom
        border = self._get_option(section, 'border', raise_=False)
        if border is not None:
            border = tuple(map(int, self._get_option(section, 'border').split()))
        else:
            # size: width, height
            # position: left, top
            size = self._get_size(section)
            if len(size) == 1:
                size.append(size[0])
            position = list(map(int, self._get_option(section, 'position', 0).split()))
            if len(position) == 1:
                position.append(position[0])
            src_size = self.buffers[src].size
            border = (position[0], position[1],
                      size[0] - src_size[0] - position[0], size[1] - src_size[1] - position[1])

        fill = self._get_option(section, 'fill', 0)
        self.buffers[dst] = ImageOps.expand(self.buffers[src], border, fill)

    def fit(self, section):
        src = self._get_option(section, 'src')
        size = self._get_size(section)
        method = self._get_option(section, 'method', Image.NEAREST)
        if method is not 0:
            method = getattr(Image, method.upper())
        bleed = self._get_option(section, 'bleed', 0, get='getfloat')
        centering = tuple(map(float, self._get_option(section, 'centering', '0.5 0.5').split()))
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = ImageOps.fit(self.buffers[src], size, method, bleed, centering)

    def flip(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = ImageOps.flip(self.buffers[src])

    def grayscale(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = ImageOps.grayscale(self.buffers[src])

    def invert(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = ImageOps.invert(self.buffers[src])

    def list_buffers(self, _):
        a = []
        for buf in sorted(self.buffers.keys()):
            a.append('%s(mode=%s size=%dx%d)' %
                     (buf, self.buffers[buf].mode, self.buffers[buf].size[0], self.buffers[buf].size[1]))
        logger.info('buffers: %s' % (', '.join(a)))

    def load(self, section):
        dst = self._get_option(section, 'dst')
        filename = self.format_str(section, self._get_option(section, 'filename'))
        self.buffers[dst] = Image.open(filename)

    def mirror(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = ImageOps.mirror(self.buffers[src])

    def merge(self, section):
        mode = self._get_mode(section)
        src1 = self._get_option(section, 'src1')
        bands = [self.buffers[src1]]
        for n in range(1, len(Image.new('mode', (1, 1)).getbands())):
            src = self._get_option(section, 'src%d' % n)
            bands.append(self.buffers[src])
        dst = self._get_option(section, 'dst', src1)
        self.buffers[dst] = Image.merge(mode, bands)

    def new(self, section):
        dst = self._get_option(section, 'dst')
        mode = self._get_option(section, 'mode')
        size = self._get_size(section)
        color = self._get_color(section, default=0)
        self.buffers[dst] = Image.new(mode, size, color)

    def paste(self, section):
        src1 = self._get_option(section, 'src1')
        src2 = self._get_option(section, 'src2')
        position = map(int, self._get_option(section, 'position', '0 0').split())
        mask = self._get_option(section, 'mask', raise_=False)
        # In keeping with most other functions allow paste() to be nondestructive.
        dst = self._get_option(section, 'dst', src1)
        if dst == src1:
            self.buffers[src1].paste(src2, position, mask)
        else:
            self.buffers[dst] = self.buffers[src1].copy()
            self.buffers[dst].paste(src2, position, mask)

    def posterize(self, section):
        src = self._get_option(section, 'src')
        bits = self._get_option(section, 'bits', get='getint')
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = ImageOps.posterize(self.buffers[src], bits)

    def resize(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        size = self._get_size(section)
        resample = self._get_option(section, 'resample', Image.NEAREST)
        if resample is not 0:
            resample = getattr(Image, resample.upper())
        self.buffers[dst] = self.buffers[src].resize(size, resample)

    def rotate(self, section):
        src = self._get_option(section, 'src')
        angle = self._get_option(section, 'angle', get='getfloat')
        resample = self._get_option(section, 'resample', Image.NEAREST)
        if resample is not 0:
            resample = getattr(Image, resample.upper())
        expand = self._get_option(section, 'expand', get='getboolean', raise_=False)
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = self.buffers[src].rotate(angle, resample, expand)

    def save(self, section):
        src = self._get_option(section, 'src')
        filename = self.format_str(section, self._get_option(section, 'filename'))
        tempfile = self._get_option(section, 'tempfile', fallback_section='common', get='getboolean', raise_=False)
        chmod = self._get_option(section, 'chmod', fallback_section='common', raise_=False)
        if chmod:
            chmod = int(chmod, 8)
        with smart_open(filename, 'wb', use_temp=tempfile, chmod=chmod) as f:
            self.buffers[src].save(f)
        logger.info('saved %s', filename)

    def split(self, section):
        src = self._get_option(section, 'src')
        images = self.buffers[src]
        for n in range(len(images)):
            dst = self._get_option(section, 'dst%d' % n)
            self.buffers[dst] = images[n]

    def solarize(self, section):
        src = self._get_option(section, 'src')
        threshold = self._get_option(section, 'bits', 128, get='getint')
        dst = self._get_option(section, 'dst', src)
        self.buffers[dst] = ImageOps.solarize(self.buffers[src], threshold)

    def symlink(self, section):
        src_filename = self.format_str(section, self._get_option(section, 'src_filename'))
        dst_filename = self.format_str(section, self._get_option(section, 'dst_filename'))
        raise_ = self._get_option(section, 'raise', True, fallback_section='common', get='getboolean')
        tempfile = self._get_option(section, 'tempfile', False, fallback_section='common', get='getboolean', raise_=False)
        try:
            if tempfile:
                atomic_symlink(src_filename, dst_filename)
            else:
                os.symlink(src_filename, dst_filename)
            logger.info('created symlink %s -> %s', src_filename, dst_filename)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            if raise_:
                raise
            else:
                logger.error('Could not create symlink %s -> %s', src_filename, dst_filename)
                logger.debug(traceback.format_exc())

    def transpose(self, section):
        src = self._get_option(section, 'src')
        dst = self._get_option(section, 'dst', src)
        method = getattr(Image, self._get_option(section, 'method'))
        self.buffers[dst] = self.buffers[src].tranpose(method)


def unescape_unicode(s):
    if sys.version_info[0] >= 3:
        return bytes(s, 'UTF-8').decode('unicode-escape')
    else:
        return s.decode('unicode-escape')

def read_config_file(filename):
    """Read config file."""
    logger.info('Reading config file ' + filename)
    config = RawConfigParser()

    config.add_section('daemon')
    config.set('daemon', 'user', 'pi')
    config.set('daemon', 'group', 'pi')

    config.add_section('upload')
    # User must add appropriate values

    if filename:
        config_files_read = config.read(filename)
        if filename not in config_files_read:
            raise UserWarning('Could not read ' + filename)
        logger.debug('Successfully read ' + ', '.join(config_files_read))
    return config


def get_config_option(config, section, option,
                      default=None,
                      fallback_section=None,
                      get=None):
    if config.has_option(section, option):
        sec = section
    elif fallback_section and config.has_option(fallback_section, option):
        sec = fallback_section
    else:
        return default

    if get is None:
        return config.get(sec, option)
    elif get == 'getfraction':
        s = config.get(sec, option)
        return Fraction(*map(int, s.split('/')))
    else:
        # For 'getboolean' etc
        return getattr(config, get)(sec, option)


def cmp_value_with_option(value, config, section, option, fallback_section='common'):
    def in_operator(a, b):
        return a in b

    def not_in_operator(a, b):
        return a not in b

    def no_op(a):
        return a

    ops = {'<': operator.lt,
           '<=': operator.le,
           '==': operator.eq,
           '>': operator.gt,
           '>=': operator.ge,
           '!=': operator.ne,
           'is': operator.is_,
           'is not': operator.is_not,
           'in': in_operator,
           'not in': not_in_operator,
           }
    op_name = get_config_option(config, section, option + '_operator',
                                default='==',
                                fallback_section=fallback_section)
    if op_name not in ops:
        raise Exception('Unknown operator (%s)' % op_name)
    if isinstance(value, bool):
        cast = bool
    elif isinstance(value, (int, np.int64)):
        cast = int
    elif isinstance(value, (float, np.float64)):
        cast = float
    else:
        # Keep as str
        cast = no_op

    option_str = get_config_option(config, section, option,
                                   fallback_section=fallback_section)
    if op_name in ['in', 'not_in']:
        conf_value = map(cast, option_str.split())
    else:
        conf_value = cast(option_str)
    return ops[op_name](value, conf_value)


def get_schedule(config, forced_schedule=None):
    """Get schedule to use.

    Allow for schedule to be overridden for testing.
    """

    t = time.time()
    info = {
        # TODO: Allow cache directory and baseurl to be overridden from config settings
        'AuroraWatchUK': AuroraWatchUK_SS(),
    }

    if forced_schedule:
        sections = [forced_schedule]
    else:
        sections = [x for x in config.sections() if (x not in ['DEFAULT', 'common', 'daemon'] and
                                                     get_config_option(config, x, 'sampling_interval') is not None)]

    for sec in sections:
        # Compute solar elevation and make it available for output. Allow for location to be customised in each section
        lat = get_config_option(config, sec, 'latitude',
                                             fallback_section='common', default=0, get='getfloat')
        lon = get_config_option(config, sec, 'longitude',
                                              fallback_section='common', default=0, get='getfloat')
        info['SolarElevation'] = get_solar_elevation(lat, lon, t)

        # Compare with solar elevation
        if config.has_option(sec, 'solar_elevation') \
            and not cmp_value_with_option(info['SolarElevation'], config, sec, 'solar_elevation',
                                          fallback_section='common'):
            continue

        # Compare with AuroraWatch
        if config.has_option(sec, 'aurorawatchuk_status') \
            and not cmp_value_with_option(info['AuroraWatchUK'].status.level, config, sec, 'aurorawatchuk_status',
                                          fallback_section='common'):
            continue

        # Test file existence
        if config.has_option(sec, 'file_exists') and not os.path.exists(config.get(sec, 'file_exists')):
            continue

        # Test file non-existence
        if config.has_option(sec, 'file_not_exist') and os.path.exists(config.get(sec, 'file_not_exist')):
            continue

        # All tests passed
        return sec, info

    if forced_schedule:
        return forced_schedule, info

    return 'common', info


def get_solar_elevation(latitude, longitude, t):
    loc = astral.Location(('', '', latitude, longitude, 'UTC', 0))
    return loc.solar_elevation(datetime.datetime.utcfromtimestamp(t))


# Each sampling action is made by a new thread. This function uses a
# lock to avoid contention for the camera. If the lock cannot be
# acquired the attempt is abandoned. The lock is released after the
# sample has been taken. This means two instances of process_schedule()
# can occur at the same time, whilst the earlier one writes data and
# possibly sends a real-time data packet over the network.
def process_tasks(camera, config, schedule, schedule_info):
    task_list = get_config_option(config, schedule, 'tasks', fallback_section='common', default='')
    logger.debug('tasks: ' + repr(task_list))
    tasks = Tasks(camera, config, schedule, schedule_info)
    tasks.run_tasks(task_list.split())
    return


logger = logging.getLogger(__name__)



from string import Formatter


__author__ = 'Steve Marple'
__version__ = '0.0.11'
__license__ = 'MIT'


class MyFormatter(Formatter):
    def __init__(self):
        Formatter.__init__(self)

    def convert_field(self, value, conversion):
        # do any conversion on the resulting object
        if conversion is None:
            return value
        elif conversion == 'a':
            return value.encode('ascii')
        elif conversion == 's':
            return str(value)
        elif conversion == 'r':
            return repr(value)
        elif conversion == 'l':
            return str(value).lower()
        elif conversion == 'u':
            return str(value).upper()
        elif conversion == 'c':
            return str(value).capitalize()
        elif conversion == 't':
            return str(value).title()
        elif conversion == '.':
            return str(value).rstrip('.')
        raise ValueError("Unknown conversion specifier {0!s}".format(conversion))


class LatLon(object):
    def __init__(self, latitude, longitude):
        self._latitude = DegMS(latitude, ('N', 'S'))
        self._longitude = DegMS(longitude, ('E', 'W'))

    @property
    def latitude(self):
        return self._latitude

    @property
    def longitude(self):
        return self._longitude


class DegMS(object):
    def __init__(self, val, hemi):
        self._val = val
        self._hemi = tuple(hemi)

    def __format__(self, *args, **kwargs):
        """Format the class as if it was a floating point value."""
        return self._val.__format__(*args, **kwargs)

    def _hemisphere(self):
        if self._val > 0:
            return self._hemi[0]
        elif self._val < 0:
            return self._hemi[1]
        else:
            return ''

    def _dmsh(self):
        m, s = divmod(abs(self._val) * 3600, 60)
        d, m = divmod(m, 60)
        # d = d if self._val >= 0 else -d
        return int(d), int(m), s, self._hemisphere()

    @property
    def unsigned_degrees(self):
        return abs(self._val)

    @property
    def signed_degrees(self):
        return self._val

    @property
    def minutes(self):
        return self._dmsh()[1]

    @property
    def seconds(self):
        return self._dmsh()[2]

    @property
    def hemisphere(self):
        return self._hemisphere()

    @property
    def dhms(self):
        return self._dmsh()

